//
//  gfx.h
//  FF6Tools
//
//  Created by Evan Owen on 12/7/14.
//  Copyright (c) 2014 Evan Owen. All rights reserved.
//

#ifndef __ff3tools__gfx__
#define __ff3tools__gfx__

#include <stdbool.h>

typedef unsigned char byte;
typedef unsigned short word;
typedef unsigned int dword;

// graphics formats

#define LINEAR_8BPP         0
#define LINEAR_2BPP         1
#define LINEAR_4BPP         2
#define NES_1BPP            3
#define NES_2BPP            4
#define SNES_2BPP           5
#define SNES_3BPP           6
#define SNES_4BPP           7
#define GENESIS_4BPP        8
#define SNES_8BPP           9
#define SNES_MODE7          10
#define SNES_MODE7_EX       11
#define SNES_MODE7_ZELDA    12
#define DEFAULT_GRAPHICS    LINEAR_8BPP

// palette formats

#define XRGB8888            100
#define RGB888              101
#define RGB565              102
#define RGB555              103
#define BGR555              104
#define RGB332              105
#define NES_PALETTE         106
#define DEFAULT_PALETTE     XRGB8888

// tile formats

#define ROMTOOLS_TILE       200 // vh--zzzz pppppppp mmmmmmmm mmmmmmmm
#define SNES_2BPP_TILE      201 // vhzpppmm mmmmmmmm
#define SNES_4BPP_TILE      202 // vhzpppmm mmmmmmmm
#define SNES_SPRITE_TILE    203 // vhzzpppm mmmmmmmm
#define SNES_MODE7_TILE     204 // mmmmmmmm
#define GBA_2BPP_TILE       205 // ppppvhmm mmmmmmmm
#define GBA_4BPP_TILE       206 // ppppvhmm mmmmmmmm
#define NES_TILE            207 // ------pp mmmmmmmm
#define GB_TILE             208 // zvh-bppp mmmmmmmm
#define DEFAULT_TILE        ROMTOOLS_TILE

// z-levels (these match snes9x)
#define Z_LEVEL_BACK_AREA   0
#define Z_LEVEL_BG4_LO      2
#define Z_LEVEL_BG3_LO      3
#define Z_LEVEL_SPRITE_0    4
#define Z_LEVEL_BG4_HI      6
#define Z_LEVEL_BG3_HI      7
#define Z_LEVEL_SPRITE_1    8
#define Z_LEVEL_BG2_LO      10
#define Z_LEVEL_BG1_LO      11
#define Z_LEVEL_SPRITE_2    12
#define Z_LEVEL_BG2_HI      14
#define Z_LEVEL_BG1_HI      15
#define Z_LEVEL_SPRITE_3    16
#define Z_LEVEL_BG3_HI_P    17
#define Z_LEVEL_MASK        18

// function declarations

typedef struct {
    struct Layer {
        int format;
        int height;
        int width;
        int v_scroll;
        int h_scroll;
        int z[16];
        const void* gfx;
        const void* tiles;
        bool main;
        bool sub;
        bool math;
    } layers[4];
    const void* pal;
    bool subtract;
    bool half;
    bool back_area;
    bool flipped;
//    struct HDMA {
//        bool enabled;
//        int ppu_reg;
//        void* data;
//        int line_counter;
//    } hdma[8];
} PPU;

void gfx_render(void* dest, const void* gfx, const void* pal, int length, int ppl);
void gfx_render_ppu(void* dest, PPU ppu, int length, int ppl);
//void gfx_render_ppu_hdma(void* dest, PPU ppu, int length, int ppl);
void gfx_decode(void* dest, const void* src, int length, int format);
void gfx_encode(void* dest, const void* src, int length, int format);

#endif /* defined(__ff3tools__gfx__) */
