//
//  ScriptLinkField.swift
//  CustomControls
//
//  Created by Evan Owen on 6/8/16.
//  Copyright © 2016 Evan Owen. All rights reserved.
//

import Cocoa

//@objc protocol ScriptLinkFieldDelegate {
//    func clickedScriptLink(scriptLinkField: ScriptLinkField)
//    func commandForReference(reference: Int, scriptIdentifier: String?) -> ROMCommand?
//}

class ScriptLinkFormatter: NSFormatter {
    
    override func stringForObjectValue(obj: AnyObject) -> String? {
        
        guard let reference = obj as? ROMReference,
            let referencedCommand = reference.referencedAssembly as? ROMCommand else {
                return "<Invalid Address>"
        }
        
        return referencedCommand.label
    }
    
    override func getObjectValue(obj: AutoreleasingUnsafeMutablePointer<AnyObject?>, forString string: String, errorDescription error: AutoreleasingUnsafeMutablePointer<NSString?>) -> Bool {
        return false
    }
}

class ScriptLinkField: NSTextField {

//    @IBOutlet var scriptDelegate: AnyObject?
    @IBOutlet weak var windowController: ROMWindowController?
    var button: NSButton!
    var scriptIdentifier: String?
    var script: ROMScript? {
        if scriptIdentifier == nil { return nil }
        return windowController?.rom.valueForKey(scriptIdentifier!) as? ROMScript
    }
    var encodingName: String?

    var buttonFrame: NSRect {
        
        // TODO: right to left reading order
        if userInterfaceLayoutDirection == .LeftToRight {
            
        } else {
            
        }
        
        switch controlSize {
        case .RegularControlSize:
            return NSMakeRect(frame.width - 16, 3.0, 16.0, 16.0)
        case .SmallControlSize:
            return NSMakeRect(frame.width - 15, 2.0, 15.0, 15.0)
        case .MiniControlSize:
            return NSMakeRect(frame.width - 12, 2.0, 12.0, 12.0)
        }
    }

    override init(frame frameRect: NSRect) {
        super.init(frame: frameRect)
        initButton()
    }
    
    required init?(coder: NSCoder) {
        super.init(coder: coder)
        initButton()
    }
    
    func initButton() {
        
        selectable = false
        let scriptLinkFormatter = ScriptLinkFormatter()
//        scriptLinkFormatter.scriptLinkField = self
        formatter = scriptLinkFormatter
        
        button = NSButton(frame: buttonFrame)
        button.setButtonType(NSButtonType.MomentaryChangeButton)
        button.bezelStyle = NSBezelStyle.InlineBezelStyle
        button.bordered = false
        button.title = ""
        button.target = self
        button.action = #selector(ScriptLinkField.buttonAction(_:))
        button.bind("hidden", toObject: self, withKeyPath: "hidden", options: nil)
        button.bind("controlSize", toObject: self, withKeyPath: "controlSize", options: nil)
        button.bind("enabled", toObject: self, withKeyPath: "enabled", options: nil)
        
        addSubview(button)
        
        registerForDraggedTypes(["com.ekowen86.romtools.ROMCommand"])
    }
    
    override func awakeFromNib() {
        let linkImage = NSImage(named: "ScriptMenuTemplate")!
        let buttonImage = NSImage(size: button.frame.size)
        buttonImage.lockFocus()
        linkImage.drawInRect(NSMakeRect(0, 0, buttonImage.size.width, buttonImage.size.height), fromRect: NSZeroRect, operation: NSCompositingOperation.CompositeSourceOver, fraction: 1.0)
        buttonImage.template = true
        buttonImage.unlockFocus()
        button.image = buttonImage
    }
    
    override func resizeSubviewsWithOldSize(oldSize: NSSize) {
        button.frame = buttonFrame
    }
    
    @IBAction func buttonAction(sender: AnyObject?) {
        
        if let reference = objectValue as? ROMReference,
            let referencedCommand = reference.referencedAssembly as? ROMCommand {
            windowController?.deselectAllObjects()
            windowController?.selectObject(referencedCommand)
        }
    }

    func commandForReference(reference: Int) -> ROMCommand? {
        return script?.references?[reference] as? ROMCommand
    }

    override func draggingEntered(sender: NSDraggingInfo) -> NSDragOperation {
        return NSDragOperation.Link
    }
    
    override func performDragOperation(sender: NSDraggingInfo) -> Bool {
        
        guard let pasteboardItem = sender.draggingPasteboard().pasteboardItems?.first,
            let encoding = pasteboardItem.stringForType("com.ekowen86.romtools.ROMCommand.encoding"),
            let referenceData = pasteboardItem.dataForType("com.ekowen86.romtools.ROMCommand.reference") else { return false }
        
        // make sure the encoding matches (use the script identifier if no encoding is specified)
        if encoding != encodingName ?? scriptIdentifier { return false }
        
        // read the reference value (8 bytes for 64-bit machines, not sure if this will work for 32-bit)
        var reference = 0
        referenceData.getBytes(&reference, length: 8)
        
        // get the referenced command
        guard let referencedCommand = commandForReference(reference) else { return false }

        // update the property bound to this control's value
        if let bindingInfo = infoForBinding(NSValueBinding),
            let observedKeyPath = bindingInfo[NSObservedKeyPathKey] as? String,
            let observedObject = bindingInfo[NSObservedObjectKey],
            let reference = observedObject.valueForKeyPath(observedKeyPath)?.copy() as? ROMReference {
            
            reference.referencedAssembly = referencedCommand

            observedObject.setValue(reference, forKeyPath: observedKeyPath)
            objectValue = reference
        }

        return true
    }
}
