//
//  FF6AttackFrameView.swift
//  FF6Tools
//
//  Created by Evan Owen on 12/15/16.
//  Copyright © 2016 Evan Owen. All rights reserved.
//

import Cocoa

class FF6AttackFrameView: ROMView {

    @IBOutlet weak var attackController: FF6AttackController!
    
    var zoom: CGFloat {
        return (frame.height - 38) / CGFloat(attackController.selectedLayer.frameHeight * 16)
    }
    
    var frameLayers = [Int : CGLayer]()
    
    override func setFrameSize(_ newSize: NSSize) {
        
        if !(superview is NSClipView) {
            super.setFrameSize(newSize)
            return
        }
        
        let width = rect(forFrame: 0).width * CGFloat(attackController.selectedLayer.frameCount) + 8
        super.setFrameSize(NSMakeSize(width, newSize.height))
    }
    
    override func resize(withOldSuperviewSize oldSize: NSSize) {
        super.resize(withOldSuperviewSize: oldSize)
        if let clipView = superview as? NSClipView {
            setFrameSize(clipView.documentVisibleRect.size)
            needsDisplay = true
        }
    }

    func rect(forFrame frameIndex: Int) -> NSRect {
        // size of one frame
        let width = CGFloat(attackController.selectedLayer.frameWidth) * 16 * zoom + 8
        let height = frame.height
        
        return NSMakeRect(CGFloat(frameIndex) * width + 4, 0, width, height)
    }

    func rect(forFrameImage frameIndex: Int) -> NSRect {
        return rect(forFrame: frameIndex).offsetBy(dx: 0, dy: 11).insetBy(dx: 4, dy: 19)
    }

    func rect(forFrameText frameIndex: Int) -> NSRect {
        
        var frameRect = rect(forFrame: frameIndex)
        frameRect.origin.x += 4
        frameRect.origin.y += 8
        frameRect.size.width -= 8
        frameRect.size.height = 14
        
        return frameRect
    }

    override func draw(_ dirtyRect: NSRect) {
        super.draw(dirtyRect)

        // fill the view with the background color
        if let clipView = superview as? NSClipView, let scrollView = clipView.superview as? NSScrollView {
            scrollView.backgroundColor = NSColor.windowBackgroundColor
        }
        NSColor.windowBackgroundColor.set()
        NSBezierPath.fill(dirtyRect)
        NSGraphicsContext.current?.imageInterpolation = imageInterpolation
        
        for frameIndex in 0..<attackController.selectedLayer.frameCount {

            if frameLayers[frameIndex] == nil {
                if let currentContext = NSGraphicsContext.current?.cgContext {
                    let frameSize = NSMakeSize(CGFloat(attackController.selectedLayer.frameWidth * 16), CGFloat(attackController.selectedLayer.frameHeight * 16))
                    frameLayers[frameIndex] = CGLayer(currentContext, size: frameSize, auxiliaryInfo: nil)
                    drawFrame(frameIndex: frameIndex)
                } else {
                    rom.log("Unable to create frame layer")
                    continue
                }
            }

//            if validFrames[frameIndex] == true { continue }
//            drawFrame(frameIndex: frameIndex)
//            validFrames[frameIndex] = true
//            continue
            
//            // load the frame data
//            guard let frameData = attackController.rom?.value(at: attackController.selectedLayer.frameOffset + frameIndex, inPropertyWithKey: "attackGraphicsFrames") as? ROMData else { continue }
//            attackController.selectedLayer.frames = frameData.data
//            attackController.selectedLayer.decodeLayout()
//            let frameImage = attackController.selectedLayer.image
//            if frameImage == nil { break }
//
//            // draw the frame image
//            let frameRect = rect(forFrameImage: frameIndex)
//            NSColor.black.set()
//            NSBezierPath.fill(frameRect)
//            NSGraphicsContext.current?.imageInterpolation = imageInterpolation
//            NSGraphicsContext.current?.cgContext.draw(frameImage!, in: frameRect)

            // draw the frame image
            let frameRect = rect(forFrameImage: frameIndex)
            NSGraphicsContext.current?.cgContext.draw(frameLayers[frameIndex]!, in: frameRect)

            // draw the frame label
            let frameText = NSString(format: "Frame %d", frameIndex)
            let style = NSMutableParagraphStyle()
            style.alignment = NSTextAlignment.center
            frameText.draw(with: rect(forFrameText: frameIndex), options: [], attributes: [NSAttributedStringKey.paragraphStyle : style])

            // draw the selection border
            if frameIndex != attackController.selectedFrame { continue }
            NSColor.keyboardFocusIndicatorColor.set()
            let borderPath = NSBezierPath(roundedRect: frameRect.insetBy(dx: -2.0, dy: -2.0), xRadius: 1.0, yRadius: 1.0)
            borderPath.lineWidth = 4.0
            borderPath.stroke()
        }
    }
    
    func drawFrame(frameIndex: Int) {
        
        // load the frame data
        guard let frameData = attackController.rom?.value(at: attackController.selectedLayer.frameOffset + frameIndex, inPropertyWithKey: "attackGraphicsFrames") as? ROMData else { return }
        attackController.selectedLayer.frames = frameData.data
        attackController.selectedLayer.decodeLayout()
        let frameImage = attackController.selectedLayer.image
        if frameImage == nil { return }
        
        // draw the frame image
        let frameRect = NSMakeRect(0, 0, CGFloat(attackController.selectedLayer.frameWidth * 16), CGFloat(attackController.selectedLayer.frameHeight * 16))
//        let frameRect = rect(forFrameImage: frameIndex)
        let frameContext = frameLayers[frameIndex]?.context
        frameContext?.setFillColor(CGColor.black)
        frameContext?.fill(frameRect)
        frameContext?.draw(frameImage!, in: frameRect)
//        NSColor.black.set()
//        NSBezierPath.fill(frameRect)
//        NSGraphicsContext.current?.imageInterpolation = imageInterpolation
//        NSGraphicsContext.current?.cgContext.draw(frameImage!, in: frameRect)
        
//        // draw the frame label
//        let frameText = NSString(format: "Frame %d", frameIndex)
//        let style = NSMutableParagraphStyle()
//        style.alignment = NSTextAlignment.center
//        frameText.draw(with: rect(forFrameText: frameIndex), options: [], attributes: [NSAttributedStringKey.paragraphStyle : style])
        
//        // draw the selection border
//        if frameIndex != attackController.selectedFrame { return }
//        NSColor.keyboardFocusIndicatorColor.set()
//        let borderPath = NSBezierPath(roundedRect: frameRect.insetBy(dx: -2.0, dy: -2.0), xRadius: 1.0, yRadius: 1.0)
//        borderPath.lineWidth = 4.0
//        borderPath.stroke()
    }
    
    override func mouseDown(with event: NSEvent) {
        let clickedPoint = convert(event.locationInWindow, from: nil)
        for frame in 0..<attackController.selectedLayer.frameCount {
            if rect(forFrame: frame).contains(clickedPoint) {
                attackController.selectedFrame = frame
                needsDisplay = true
                attackController.frameEditor.needsDisplay = true
                break
            }
        }
    }
    
    override func moveRight(_ sender: Any?) {
        if attackController.selectedFrame == (attackController.selectedLayer.frameCount - 1) { return }
        attackController.selectedFrame += 1
        scrollToVisible(rect(forFrame: attackController.selectedFrame))
        needsDisplay = true
        attackController.frameEditor.needsDisplay = true
    }
    
    override func moveLeft(_ sender: Any?) {
        if attackController.selectedFrame == 0 { return }
        attackController.selectedFrame -= 1
        scrollToVisible(rect(forFrame: attackController.selectedFrame))
        needsDisplay = true
        attackController.frameEditor.needsDisplay = true
    }
    
    override var acceptsFirstResponder: Bool { return true }
}
