//
//  FF6BattleCharacter.swift
//  FF3Tools
//
//  Created by Evan Owen on 3/8/16.
//  Copyright © 2016 Evan Owen. All rights reserved.
//

import Cocoa

class FF6BattleCharacter: NSObject, ROMSelectable {
    
    // MARK: ROMSelectable properties
    
    // identifier to specify the view for the propertiesController to load
    var identifier: String? { return "battleCharacter" }
//    var properties: NSDictionary? { return ["character" : 0, "hidden" : 0, "graphics" : 0, "enemy" : 0, "aiScript" : 0, "slot" : 0, "hasAI" : 0] }
    var properties: NSDictionary? { return ["character" : 0, "hidden" : 0, "graphics" : 0, "enemy" : 0, "aiScript" : 0, "slot" : 0, "hasAI" : 0, "enableSlot1" : 0, "enableSlot2" : 0, "enableSlot3" : 0, "enableSlot4" : 0] }
    var nib: String? { return "BattleProperties" }

    var rom: ROM!
    
    var auxBattleProperties: ROMData!
    var characterAIProperties: ROMData!
    var characterSlotData: ROMData {
        return characterAIProperties.mutableArrayValue(forKey: "characters").object(at: slot - 1) as! ROMData
    }
    @objc dynamic var slot = 1

    @objc dynamic var character: Int {
        get { return characterSlotData.intValue(forKey: "character") }
        set {
            stopObservingProperties()
            characterSlotData.setValue(newValue, forKey: "character")
            startObservingProperties()
        }
    }
    @objc var hidden: Bool {
        get { return characterSlotData.boolValue(forKey: "hidden") }
        set {
            stopObservingProperties()
            characterSlotData.setValue(newValue ? 1 : 0, forKey: "hidden")
            startObservingProperties()
        }
    }
    @objc var enemy: Bool {
        get { return characterSlotData.boolValue(forKey: "enemy") }
        set {
            stopObservingProperties()
            characterSlotData.setValue(newValue ? 1 : 0, forKey: "enemy")
            startObservingProperties()
        }
    }
    @objc dynamic var graphics: Int {
        get { return characterSlotData.intValue(forKey: "graphics") }
        set {
            stopObservingProperties()
            characterSlotData.setValue(newValue, forKey: "graphics")
            startObservingProperties()
        }
    }
    @objc var aiScript: Int {
        get { return characterSlotData.intValue(forKey: "aiScript") }
        set {
            stopObservingProperties()
            characterSlotData.setValue(newValue, forKey: "aiScript")
            startObservingProperties()
        }
    }
    @objc dynamic var x: Int {
        get { return characterSlotData.intValue(forKey: "x") }
        set { characterSlotData.setValue(newValue, forKey: "x") }
    }
    @objc dynamic var y: Int {
        get { return characterSlotData.intValue(forKey: "y") }
        set { characterSlotData.setValue(newValue, forKey: "y") }
    }

    var hasAI: Bool {
        let enableCharacterAI = auxBattleProperties.boolValue(forKey: "enableCharacterAI")
        if !enableCharacterAI {
            // character AI is disabled for this battle
            return false
        } else if (character == 0x3F) && hidden && enemy { // if byte 0 == 0xFF
            // character slot is disabled
            return false
        }
        return true
    }
    
    @objc var enableSlot1: Bool {
        get { return isSlotEnabled(slot: 1) }
        set { setSlotEnabled(slot: 1, enabled: newValue) }
    }

    @objc var enableSlot2: Bool {
        get { return isSlotEnabled(slot: 2) }
        set { setSlotEnabled(slot: 2, enabled: newValue) }
    }

    @objc var enableSlot3: Bool {
        get { return isSlotEnabled(slot: 3) }
        set { setSlotEnabled(slot: 3, enabled: newValue) }
    }

    @objc var enableSlot4: Bool {
        get { return isSlotEnabled(slot: 4) }
        set { setSlotEnabled(slot: 4, enabled: newValue) }
    }

    func isSlotEnabled(slot: Int) -> Bool {
        let slot = characterAIProperties.mutableArrayValue(forKey: "characters").object(at: slot - 1) as! ROMData
        let character = slot.intValue(forKey: "character")
        let present = slot.boolValue(forKey: "hidden")
        let enemy = slot.boolValue(forKey: "enemy")
        return !((character == 0x3F) && present && enemy) // if byte 0 != 0xFF
    }
    
    func setSlotEnabled(slot: Int, enabled: Bool) {
        let slot = characterAIProperties.mutableArrayValue(forKey: "characters").object(at: slot - 1) as! ROMData
        if enabled == true {
            slot.setValue(0, forKey: "character")
            slot.setValue(0, forKey: "hidden")
            slot.setValue(0, forKey: "enemy")
//            slot.setValue(0, forKey: "x")
//            slot.setValue(0, forKey: "y")
        } else {
            slot.setValue(0x3F, forKey: "character")
            slot.setValue(1, forKey: "hidden")
            slot.setValue(1, forKey: "enemy")
        }
    }

    init(rom: ROM, auxBattleProperties: ROMData, slot: Int) {
        self.rom = rom
        self.auxBattleProperties = auxBattleProperties
        self.slot = slot
        let characterAI = auxBattleProperties.intValue(forKey: "characterAI")
        self.characterAIProperties = rom.value(at: characterAI, inPropertyWithKey: "characterAIProperties") as? ROMData

        super.init()
        
        startObservingProperties()
    }
    
    deinit {
        stopObservingProperties()
    }

    func startObservingProperties() {
        characterSlotData.addObserver(self, forKeyPath: "character", options: NSKeyValueObservingOptions.prior, context: &FF6BattleCharacter.context)
        characterSlotData.addObserver(self, forKeyPath: "hidden", options: NSKeyValueObservingOptions.prior, context: &FF6BattleCharacter.context)
        characterSlotData.addObserver(self, forKeyPath: "graphics", options: NSKeyValueObservingOptions.prior, context: &FF6BattleCharacter.context)
        characterSlotData.addObserver(self, forKeyPath: "enemy", options: NSKeyValueObservingOptions.prior, context: &FF6BattleCharacter.context)
        characterSlotData.addObserver(self, forKeyPath: "aiScript", options: NSKeyValueObservingOptions.prior, context: &FF6BattleCharacter.context)
    }
    
    func stopObservingProperties() {
        characterSlotData.removeObserver(self, forKeyPath: "character")
        characterSlotData.removeObserver(self, forKeyPath: "hidden")
        characterSlotData.removeObserver(self, forKeyPath: "graphics")
        characterSlotData.removeObserver(self, forKeyPath: "enemy")
        characterSlotData.removeObserver(self, forKeyPath: "aiScript")
    }
    
    static var context = 0
    
    override func observeValue(forKeyPath keyPath: String?, of object: Any?, change: [NSKeyValueChangeKey : Any]?, context: UnsafeMutableRawPointer?) {
        if context != &FF6BattleCharacter.context {
            super.observeValue(forKeyPath: keyPath, of: object, change: change, context: context)
            return
        }
        
        let changeFunc: (String)->() = (change?[NSKeyValueChangeKey.notificationIsPriorKey] != nil) ? willChangeValue : didChangeValue
        changeFunc(keyPath!)
        
        if ["character", "hidden", "enemy"].contains(keyPath!) {
            changeFunc("enableSlot\(slot)")
//            changeFunc("enableSlot2")
//            changeFunc("enableSlot3")
//            changeFunc("enableSlot4")
        }
    }
}
