//
//  FF6ColosseumView.swift
//  FF6Tools
//
//  Created by Evan Owen on 3/19/17.
//  Copyright © 2017 Evan Owen. All rights reserved.
//

import Cocoa

class FF6ColosseumView: FF6BattleView {

    @objc var monsterIndexKeyPath: String?

    @objc var monsterIndex = 0 {
        didSet {
            reloadMonsters()
            reloadCharacters()
            needsDisplay = true
        }
    }
    
    // force colosseum battle background
    override var background: Int { get { return 0x1C} set { } }
    
    override func awakeFromNib() {
        
        super.awakeFromNib()
        
        if monsterIndexKeyPath != nil {
            bind(NSBindingName(rawValue: "monsterIndex"), to: windowController, withKeyPath: monsterIndexKeyPath!, options: nil)
        }
    }
    
    override func reloadMonsters() {
        
        monsters.removeAll()
        guard let battleProperties = rom.value(forKey: "battleProperties") as? ROMArray,
            let dummyBattle = battleProperties.createNewAssembly() as? ROMData else { return }
        
        let dummyMonster = FF6BattleMonster(rom: ff6ROM, battleProperties: dummyBattle, slot: 1)
        dummyMonster.index = monsterIndex
        dummyMonster.present = true
        dummyMonster.x = 4
        dummyMonster.y = 4
        monsters.append(dummyMonster)

    }
    
    override func reloadCharacters() {
        characters.removeAll()
        guard let auxBattleProperties = rom.value(forKey: "auxBattleProperties") as? ROMArray,
            let dummyBattle = auxBattleProperties.createNewAssembly() as? ROMData else { return }
        
        let dummyCharacter = FF6BattleCharacter(rom: ff6ROM, auxBattleProperties: dummyBattle, slot: 1)
        characters.append(dummyCharacter)
    }
    
    override func rectForMonster(_ monster: FF6BattleMonster) -> NSRect {
        guard let alignment = rom.value(at: monster.index, inPropertyWithKey: "monsterVerticalAlignments") as? ROMData else { return NSZeroRect }
        
        // see C1/1588
        
        let monsterSize = monster.size
        let x = (128 - monsterSize.width * 8) / 2 + 24
        var y = 128
        
        if !ff6ROM.isGBA {
            switch alignment.intValue(forKey: "alignment") {
            case 0: y = monsterSize.height * 8
            case 1: break
            case 2: y -= 8
            case 3: y += 8
            case 4: y -= 40
            default: break
            }
        } else {
            switch alignment.intValue(forKey: "alignment") {
            case 0: y = monsterSize.height * 8 + 31
            case 1: break
            case 2: y -= 24
            case 3: y += 8
            case 4: break
            case 5: y += 16
            default: break
            }
        }

        return NSMakeRect(CGFloat(x - 8) * zoom, bounds.height - CGFloat(y - 4) * zoom, CGFloat(monsterSize.width * 8) * zoom, CGFloat(monsterSize.height * 8) * zoom)
    }
    
    override func rectForCharacter(_ character: FF6BattleCharacter) -> NSRect {
        
        return NSRect(x: 203 * zoom, y: bounds.height - 124 * zoom, width: 16 * zoom, height: 32 * zoom)
    }
}
