//
//  LinkPopUpButton.swift
//  CustomControls
//
//  Created by Evan Owen on 5/25/16.
//  Copyright © 2016 Evan Owen. All rights reserved.
//

import Cocoa

class LinkPopUpButton: NSPopUpButton {

    var button: NSButton!
    @objc var linkIdentifier: String?
    @objc var stringTable: String?
    @IBOutlet var windowController: ROMWindowController?

    var buttonFrame: NSRect {

        // TODO: right to left reading order
        if userInterfaceLayoutDirection == .leftToRight {
            
        } else {
            
        }
        
        switch controlSize {
        case .regular:
            return NSMakeRect(frame.width - 37, 5.0, 15.0, 15.0)
        case .small:
            return NSMakeRect(frame.width - 34, 3.0, 14.0, 14.0)
        case .mini:
            return NSMakeRect(frame.width - 28, 2.0, 11.0, 11.0)
        }
    }
    
    func initButton() {
        
        if linkIdentifier == nil { return }
        
        button = NSButton(frame: buttonFrame)
        button.setButtonType(NSButton.ButtonType.momentaryChange)
        button.bezelStyle = NSButton.BezelStyle.inline
        button.isBordered = false
        button.title = ""
        button.target = self
        button.action = #selector(LinkPopUpButton.buttonAction(_:))
        button.bind(NSBindingName(rawValue: "hidden"), to: self, withKeyPath: "hidden", options: nil)
        button.bind(NSBindingName(rawValue: "controlSize"), to: self, withKeyPath: "controlSize", options: nil)
        button.bind(NSBindingName(rawValue: "enabled"), to: self, withKeyPath: "enabled", options: nil)

        let linkImage = NSImage(named: NSImage.Name.followLinkFreestandingTemplate)!
        let buttonImage = NSImage(size: button.frame.size)
        let buttonRect = NSMakeRect(0, 0, buttonImage.size.width, buttonImage.size.height)
        buttonImage.lockFocus()
        linkImage.draw(in: buttonRect, from: NSZeroRect, operation: NSCompositingOperation.sourceOver, fraction: 1.0)
        buttonImage.isTemplate = true
        buttonImage.unlockFocus()
        button.image = buttonImage
        
        addSubview(button)
    }
    
    override func awakeFromNib() {
        
        initButton()
        
        // check if the button has a linked string table
        if stringTable == nil { return }
        guard let linkedStringTable = windowController?.rom.stringTable(forKey: stringTable!) else { return }
        
        // bind an array controller to the string table and bind this button to the array controller
        let arrayController = NSArrayController()
        arrayController.bind(NSBindingName(rawValue: "contentArray"), to: linkedStringTable, withKeyPath: "stringArray", options: nil)
        bind(NSBindingName(rawValue: "content"), to: arrayController, withKeyPath: "arrangedObjects", options: nil)
        bind(NSBindingName(rawValue: "contentObjects"), to: arrayController, withKeyPath: "arrangedObjects.index", options: nil)
        bind(NSBindingName(rawValue: "contentValues"), to: arrayController, withKeyPath: "arrangedObjects.numberedString", options: nil)
    }

    override func resizeSubviews(withOldSize oldSize: NSSize) {
        button.frame = buttonFrame
    }
    
    @IBAction func buttonAction(_ sender: AnyObject?) {
        
        // get the link identifier and the selected index
        guard let identifier = linkIdentifier else { return }
        guard let selectedIndex = selectedItem?.representedObject as? Int else { return }
        
        // select the linked object
        windowController?.deselectAllObjects()
        windowController?.selectObject(windowController?.rom.value(at: selectedIndex, inPropertyWithKey: identifier) as AnyObject?)
    }
}
