//
//  ROMTableController.swift
//  FF6Tools
//
//  Created by Evan Owen on 8/7/16.
//  Copyright © 2016 Evan Owen. All rights reserved.
//

// TODO: this is a mess - there are way too many classes here
// TODO: somehow make the table view data source observe properties so it can update on undo

import Cocoa

class ROMPropertyTableView: NoScrollView, NSTableViewDataSource {
    @IBOutlet weak var windowController: ROMWindowController?

    @objc var count: NSNumber?
    
    override func awakeFromNib() {
        guard let tableView = documentView as? NSTableView else { return }
        for object in tableView.tableColumns {
            guard let column = object as? ROMTableColumn,
                let keyPath = column.keyPath,
                let assemblyIdentifier = column.assembly else { continue }
            
            windowController?.addObserver(self, forKeyPath: assemblyIdentifier + "." + keyPath, options: NSKeyValueObservingOptions.old, context: &ROMPropertyTableView.context)
        }
    }
    
    static var context = 0
    
    override func observeValue(forKeyPath keyPath: String?, of object: Any?, change: [NSKeyValueChangeKey : Any]?, context: UnsafeMutableRawPointer?) {
        if context != &ROMPropertyTableView.context {
            super.observeValue(forKeyPath: keyPath, of: object, change: change, context: context)
        }
        
        guard let tableView = documentView as? NSTableView else { return }
        tableView.reloadData()
    }

    func numberOfRows(in tableView: NSTableView) -> Int {
        if count != nil { return Int(truncating: count!) }
        return 0
    }

    func tableView(_ tableView: NSTableView, objectValueFor tableColumn: NSTableColumn?, row: Int) -> Any? {
        guard let column = tableColumn as? ROMTableColumn else { return nil }
        
        if let stringTable = column.stringTable,
            let string = windowController?.rom.string(atIndex: row, inStringTable: stringTable) {
            return string
            
        } else if let keyPath = column.keyPath,
            let assemblyIdentifier = column.assembly,
            let flags = windowController?.value(forKeyPath: assemblyIdentifier + "." + keyPath) as? Int {
            return (flags >> row) & 1 == 1 ? true : false
        }
        return nil
    }

    func tableView(_ tableView: NSTableView, setObjectValue object: Any?, for tableColumn: NSTableColumn?, row: Int) {
        guard let column = tableColumn as? ROMTableColumn,
            let keyPath = column.keyPath,
            let assemblyIdentifier = column.assembly,
            var flags = windowController?.value(forKeyPath: assemblyIdentifier + "." + keyPath) as? Int else { return }
        
        if (object! as AnyObject).intValue == 1 {
            flags |= 1 << row
        } else {
            flags &= ~(1 << row)
        }
        windowController?.setValue(flags, forKeyPath: assemblyIdentifier + "." + keyPath)
    }
}

class ROMTableColumn: NSTableColumn {
    @objc var assembly: String?
    @objc var keyPath: String?
    @objc var stringTable: String?
}

class ROMTableView: NSScrollView {
    var assemblyIdentifier: String?
}

class ROMTableController: NSObject, NSTableViewDelegate, NSTableViewDataSource {

    @IBOutlet weak var windowController: ROMWindowController!
    var rom: ROM { return windowController.rom }

    @IBOutlet var tableView: NSTableView!
    @IBOutlet var parentView: NSView!

    fileprivate var selectedArray: AnyObject?
//    fileprivate var selectedStringTable: ROMStringTable?
    
    override func awakeFromNib() {
        tableView.action = #selector(ROMTableController.clickedRow(_:))
        tableView.target = self
    }

    func selectObject(_ object: AnyObject?) {
        
        if selectedArray === object { return }
        selectedArray = object
        tableView.reloadData()
        tableView.scrollRowToVisible(0)
        tableView.deselectAll(self)
    }
    
    @IBAction func add(_ sender: AnyObject) {
        
        if let romArray = sender as? ROMArray ?? selectedArray as? ROMArray,
            let newAssembly = romArray.createNewAssembly() {
            
            // append new entries to linked arrays first
            if let linkedArrays = romArray.linkedArrays {
                for object in linkedArrays {
                    guard let arrayIdentifier = object as? String else { continue }
                    // skip linked arrays if the count doesn't match
                    if let linkedArray = rom.value(forKey: arrayIdentifier) as? ROMArray, linkedArray.count == romArray.count {
                        add(linkedArray)
                    }
                }
            }
            
            // append the new assembly to the array
            insertObject(newAssembly, inArray: romArray, atIndex: romArray.count)

            // append a new string to the string table
            if let identifier = romArray.identifier,
                let stringTable = rom.stringTable(forKey: identifier) {
                let newString = stringTable.createNewStringAtIndex(romArray.count - 1)
                insertObject(newString, inArray: stringTable, atIndex: romArray.count - 1)
            }
            
            if romArray === selectedArray {
                tableView.reloadData()
                tableView.selectRowIndexes(IndexSet(integer: romArray.count - 1), byExtendingSelection: false)
                tableView.scrollRowToVisible(romArray.count - 1)
            }
        }
    }

    @IBAction func delete(_ sender: AnyObject) {
        
        if let romArray = sender as? ROMArray ?? selectedArray as? ROMArray,
            let assembly = romArray.array.lastObject as? ROMAssembly {

            // remove entries from linked arrays first
            if let linkedArrays = romArray.linkedArrays {
                for object in linkedArrays {
                    guard let arrayIdentifier = object as? String else { continue }
                    // skip linked arrays if the count doesn't match
                    if let linkedArray = rom.value(forKey: arrayIdentifier) as? ROMArray , linkedArray.count == romArray.count {
                        delete(linkedArray)
                    }
                }
            }
            
            // remove the associated string from the string table
            if let identifier = romArray.identifier,
                let stringTable = rom.stringTable(forKey: identifier),
                let string = stringTable[romArray.count - 1] as? ROMString {
                removeObject(string, fromArray: stringTable)
            }
            
            // remove the last assembly from the array
            removeObject(assembly, fromArray: romArray)
            
        }
    }
    
    @objc func insertObject(_ object: AnyObject, inArray array: AnyObject?, atIndex index: Int) {
        
        if let romArray = array as? ROMArray, let assembly = object as? ROMAssembly {
            romArray.insertAssembly(assembly, atIndex: index)
            
            // save undo data
            if let undo = windowController?.undoManager {
                let target = undo.prepare(withInvocationTarget: self) as AnyObject
                target.removeObject(assembly, fromArray: romArray)
                if (!undo.isUndoing) {
                    undo.setActionName("Insert Assembly")
                }
            }
            
        } else if let stringTable = array as? ROMStringTable,
            let string = object as? ROMString {
            stringTable.addString(string)
            
            // save undo data
            if let undo = windowController?.undoManager {
                let target = undo.prepare(withInvocationTarget: self) as AnyObject
                target.removeObject(string, fromArray: stringTable)
                if (!undo.isUndoing) {
                    undo.setActionName("Insert String")
                }
            }
        }
        
        if array === selectedArray {
            tableView.reloadData()
            tableView.scrollRowToVisible(index)
        }
    }
    
    @objc func removeObject(_ object: AnyObject, fromArray array: AnyObject?) {
        
        windowController.deselectAllObjects()
        
        if let romArray = array as? ROMArray, let assembly = object as? ROMAssembly {
            
            let index = romArray.array.index(of: assembly)
            if index == NSNotFound { return }
            
            // save undo data
            if let undo = windowController?.undoManager {
                let target = undo.prepare(withInvocationTarget: self) as AnyObject
                target.insertObject(assembly, inArray: romArray, atIndex: index)
                if (!undo.isUndoing) {
                    undo.setActionName("Remove Assembly")
                }
            }
            
            romArray.removeAssemblyAtIndex(index)
            
        } else if let stringTable = array as? ROMStringTable, let string = object as? ROMString {
            
            if stringTable[string.index] !== string { return }
            
            // save undo data
            if let undo = windowController?.undoManager {
                let target = undo.prepare(withInvocationTarget: self) as AnyObject
                target.insertObject(string, inArray: stringTable, atIndex: string.index)
                if (!undo.isUndoing) {
                    undo.setActionName("Remove String")
                }
            }
            
            stringTable.removeString(string)
        }
        
        if array === selectedArray {
            windowController.deselectAllObjects()
            tableView.reloadData()
        }

    }
    
    // MARK: NSKeyValueObserving methods
    
    static var context = 0
    
    fileprivate var selectedObject: AnyObject?
    
    func selectObjectAtIndex(index: Int, inArray array: AnyObject?) {
        
        // remove observers for the previous selected object
        if selectedObject is ROMText {
            selectedObject?.removeObserver(self, forKeyPath: "string", context: &ROMTableController.context)
        } else if selectedObject is ROMString {
            selectedObject?.removeObserver(self, forKeyPath: "formattedString", context: &ROMTableController.context)
        }
        selectedObject = nil
        
        // select the object at the selected row
        windowController.deselectAllObjects()
//        windowController.selectObject(atIndex: index, inArray: array)
        
        // set the new selected object
//        if let stringTable = selectedStringTable,
//            let string = stringTable[index] as? ROMString {
//            
//            // linked string table
//            selectedObject = string
//            if let selected = array as? ROMSelectable, let identifier = selected.identifier {
//                windowController.selectObject(string, identifierOverride: identifier + "-string")
//            } else {
//                windowController.selectObject(string)
//            }
        
        if let romArray = selectedArray as? ROMArray,
            let selectedObject = romArray[index] as? ROMSelectable {
            windowController.selectObject(selectedObject)
        } else if let stringTable = selectedArray as? ROMStringTable,
            let selectedString = stringTable[index] as? ROMString {
            windowController.selectObject(selectedString)
        }

        if let romArray = selectedArray as? ROMArray,
            let stringTableIdentifier = romArray.linkedStringTable,
            let stringTable = rom.stringTable(forKey: stringTableIdentifier),
            let string = stringTable[index] as? ROMString {
            
            // linked string table
            selectedObject = string

        } else if let romArray = selectedArray as? ROMArray,
            let text = romArray[index] as? ROMText {
            
            // formatted ROMText
            selectedObject = text
            
        } else if let stringTable = selectedArray as? ROMStringTable,
            let string = stringTable[index] as? ROMString {
            
            // string table value
            selectedObject = string
        }

        // add observers for the string that is displayed in the list
        if selectedObject is ROMText {
            selectedObject?.addObserver(self, forKeyPath: "string", options: NSKeyValueObservingOptions.old, context: &ROMTableController.context)
        } else if selectedObject is ROMString {
            selectedObject?.addObserver(self, forKeyPath: "formattedString", options: NSKeyValueObservingOptions.old, context: &ROMTableController.context)
        }
    }
    
    override func observeValue(forKeyPath keyPath: String?, of object: Any?, change: [NSKeyValueChangeKey : Any]?, context: UnsafeMutableRawPointer?) {
        if context != &ROMTableController.context {
            super.observeValue(forKeyPath: keyPath, of: object, change: change, context: context)
            return
        }
        
        // reload the selected row
        if tableView.selectedRow == -1 { return }
        tableView.reloadData(forRowIndexes: IndexSet(integer: tableView.selectedRow), columnIndexes: IndexSet(integer: tableView.column(withIdentifier: NSUserInterfaceItemIdentifier(rawValue:"name"))))
        tableView.noteHeightOfRows(withIndexesChanged: IndexSet(integer: tableView.selectedRow))
    }
    
    // MARK: NSTableViewDelegate methods
    
    func numberOfRows(in tableView: NSTableView) -> Int {
        
        if let romArray = selectedArray as? ROMArray {
            return romArray.count
        } else if let stringTable = selectedArray as? ROMStringTable {
            return stringTable.stringArray.count
        }
        
        return 1
    }
    
    func tableView(_ tableView: NSTableView, objectValueFor tableColumn: NSTableColumn?, row: Int) -> Any? {
        
        if selectedArray == nil {
            return "No Selection"
            
        } else if (tableColumn?.identifier)!.rawValue == "index" {
            return row
            
        } else if (tableColumn?.identifier)!.rawValue == "hexIndex" {
            return String(format: "0x%04X", row)
            
        } else if (tableColumn?.identifier)!.rawValue == "name" {
            
//            if let stringTable = selectedStringTable,
//                let string = stringTable[row] as? ROMString {
//                
//                // linked string table
//                return string.formattedString

            if let romArray = selectedArray as? ROMArray,
                let stringTableIdentifier = romArray.linkedStringTable,
                let stringTable = rom.stringTable(forKey: stringTableIdentifier),
                let string = stringTable[row] as? ROMString {
                
                // linked string table
                return string.formattedString
                
            } else if let romArray = selectedArray as? ROMArray,
                let text = romArray[row] as? ROMText {
                // formatted ROMText
                
                var textString = text.formattedString()
                
                textString = textString.replacingOccurrences(of: "\u{12}", with: "\n")
                
                // replace multiple choice indicators
                var choiceIndex = 1
                while textString.contains("\u{5}") {
                    textString.replaceSubrange(textString.range(of: "\u{5}")!, with: String(choiceIndex) + ".")
                    choiceIndex += 1
                }
                return textString
                
            } else if let stringTable = selectedArray as? ROMStringTable,
                let string = stringTable[row] as? ROMString {
                
                // string table value
                return string.formattedString
            }
        }
       
        return String(format: "Row %d", row)
    }
    
    func tableView(_ tableView: NSTableView, heightOfRow row: Int) -> CGFloat {
        if let romArray = selectedArray as? ROMArray, row < romArray.count, romArray.linkedStringTable == nil {
            if let text = romArray[row] as? ROMText {
                return CGFloat(text.numLines) * tableView.rowHeight
            }
        }
        return tableView.rowHeight
    }
    
    @objc func clickedRow(_ sender: AnyObject) {
        
        if tableView.clickedRow == -1 { return }
        
        // this normally does nothing and selection is handled by tableViewSelectionDidChange:
        // if the selected row in the table view is clicked and it's not selected in the window controller, this function selects it 
        
        if let romArray = selectedArray as? ROMArray, tableView.clickedRow < romArray.count,
            !windowController.objectIsSelected(romArray[tableView.clickedRow]) {
            selectObjectAtIndex(index: tableView.clickedRow, inArray: selectedArray)
        } else if let stringTable = selectedArray as? ROMStringTable,
            !windowController.objectIsSelected(stringTable[tableView.clickedRow]) {
            selectObjectAtIndex(index: tableView.clickedRow, inArray: selectedArray)
        }
    }
    
    func tableViewSelectionDidChange(_ notification: Notification) {
        
        guard let tableView = notification.object as? NSTableView else { return }
        
        if tableView.selectedRow == -1 { return }
        
        if (selectedArray is ROMArray) || (selectedArray is ROMStringTable) {
            selectObjectAtIndex(index: tableView.selectedRow, inArray: selectedArray)
        }
    }
    
    // MARK: search methods
    
    @IBAction func search(_ sender: AnyObject) {
        
        guard let searchField = sender as? NSSearchField else { return }
        let searchString = searchField.stringValue
        if searchString.isEmpty { return }
        
        let columnIndex = tableView.column(withIdentifier: NSUserInterfaceItemIdentifier(rawValue: "name"))
        if columnIndex == -1 { return }
        let searchColumn = tableView.tableColumns[columnIndex]
        
        // start at the top if nothing is selected
        let startRow = tableView.selectedRow == -1 ? 0 : tableView.selectedRow

        var searchRow = startRow
        repeat {
            searchRow += 1
            if searchRow >= tableView.numberOfRows { searchRow = 0 }
            
            guard let rowString = tableView(tableView, objectValueFor: searchColumn, row: searchRow) as? String else { continue }
            if rowString.localizedCaseInsensitiveContains(searchString) {
                tableView.selectRowIndexes(IndexSet(integer: searchRow), byExtendingSelection: false)
                tableView.scrollRowToVisible(searchRow)
                return
            }
        } while searchRow != startRow
        
        // search string not found
        NSSound.beep()
    }
}
