//
//  ROMTileLayout.swift
//  FF6Tools
//
//  Created by Evan Owen on 6/4/17.
//  Copyright © 2017 Evan Owen. All rights reserved.
//

import Foundation

enum ROMTileFormat: String {
    case SNES2bppTile =         "SNES2bppTile"
    case SNES4bppTile =         "SNES4bppTile"
    case SNES8bppMode7 =        "SNES8bppMode7"
    case SNES4bppMode7 =        "SNES4bppMode7"
    case Default =              "Default"
    case GBA2bppTile =          "GBA2bppTile"
    case GBA4bppTile =          "GBA4bppTile"
    
    var rawFormat: Int32 {
        switch self {
        case .SNES2bppTile:         return SNES_2BPP_TILE
        case .SNES4bppTile:         return SNES_4BPP_TILE
        case .SNES8bppMode7:        return SNES_MODE7_TILE
        case .SNES4bppMode7:        return SNES_MODE7_TILE
        case .GBA2bppTile:          return GBA_2BPP_TILE
        case .GBA4bppTile:          return GBA_4BPP_TILE
        case .Default:              return DEFAULT_TILE
        }
    }
    
    var bytesPerTile: Int {
        switch self {
        case .SNES2bppTile, .SNES4bppTile, .GBA2bppTile, .GBA4bppTile:
            return 2
        case .SNES8bppMode7, .SNES4bppMode7:
            return 1
        case .Default:
            return 4
        }
    }
    
    var colorsPerPalette: Int {
        switch self {
        case .SNES2bppTile, .GBA2bppTile:
            return 4
        case .SNES4bppTile, .SNES8bppMode7, .SNES4bppMode7, .GBA4bppTile:
            return 16
        case .Default:
            return 1
        }
    }
}

class ROMTileLayout: ROMData {
    
    var width: Int
    var height: Int
    var tileFormat: ROMTileFormat = .Default
    var graphics: NSDictionary
    var palette: NSDictionary
    var paletteAssignment: String?
    var interlace: String?
    
    override init(_ rom: ROM, definition: NSDictionary?) {
        width = Int(object: definition, keyPath: "width") ?? 0
        height = Int(object: definition, keyPath: "height") ?? 0
        if let formatName = definition?["format"] as? String,
            let tileFormat = ROMTileFormat(rawValue: formatName) {
            self.tileFormat = tileFormat
        }
        graphics = definition?["graphics"] as? NSDictionary ?? [:]
        palette = definition?["palette"] as? NSDictionary ?? [:]
        paletteAssignment = definition?["paletteAssignment"] as? String
        interlace = definition?["interlace"] as? String
        
        super.init(rom, definition: definition)
    }
    
    override func createDefinition() -> NSMutableDictionary {
        let definition = super.createDefinition()
        
        definition["type"] = "ROMTileLayout"
        definition["width"] = width
        definition["height"] = height
        definition["format"] = tileFormat.rawValue
        definition["graphics"] = graphics
        definition["palette"] = palette
        definition["paletteAssignment"] = paletteAssignment
        definition["interlace"] = interlace
        
        return definition
    }
}
