//
//  TextStepperField.swift
//  CustomControls
//
//  Created by Evan Owen on 5/23/16.
//  Copyright © 2016 Evan Owen. All rights reserved.
//

import Cocoa

class TextStepperField: NSTextField, NSTextFieldDelegate {

    var stepper: NSStepper!
    
    var stepperWidth: CGFloat {
        switch controlSize {
        case .regular:
            return 16.0
        case .small:
            return 14.0
        case .mini:
            return 12.0
        }
    }
    var smallIncrement = 1 {
        didSet {
            stepper.increment = Double(smallIncrement)
        }
    }
    var bigIncrement = 10
    
    override init(frame frameRect: NSRect) {
        super.init(frame: frameRect)
        initStepper()
    }
    
    required init?(coder: NSCoder) {
        super.init(coder: coder)
    }
    
    override func awakeFromNib() {
        initStepper()
    }
    
    func initStepper() {
        
        // TODO: right to left reading order
        if userInterfaceLayoutDirection == .leftToRight {
            
        } else {
            
        }
        
//        integerValue = 0
        delegate = self

        let stepperFrame = NSMakeRect(frame.width - stepperWidth, 0, stepperWidth, frame.height)
        
        stepper = NSStepper(frame: stepperFrame)
        stepper.integerValue = integerValue
        stepper.bind(NSBindingName(rawValue: "hidden"), to: self, withKeyPath: "hidden", options: nil)
        stepper.bind(NSBindingName(rawValue: "controlSize"), to: self, withKeyPath: "controlSize", options: nil)
        stepper.bind(NSBindingName(rawValue: "enabled"), to: self, withKeyPath: "enabled", options: nil)
        if let numberFormatter = self.formatter as? NumberFormatter, let max = numberFormatter.maximum, let min = numberFormatter.minimum {
            stepper.maxValue = Double(truncating: max)
            stepper.minValue = Double(truncating: min)
        }
        stepper.target = self
        stepper.action = #selector(TextStepperField.stepperAction(_:))
        stepper.valueWraps = false
        stepper.autorepeat = true
        stepper.increment = Double(smallIncrement)
        
        addSubview(stepper)
        if let bindingInfo = infoForBinding(NSBindingName.value),
            let observedObject = bindingInfo[NSBindingInfoKey.observedObject],
            let observedKeyPath = bindingInfo[NSBindingInfoKey.observedKeyPath] as? String {
            stepper.bind(NSBindingName.value, to: observedObject, withKeyPath: observedKeyPath, options: nil)
//        } else {
        }
    }
    
    @objc func stepperAction(_ sender: AnyObject) {
        window?.makeFirstResponder(self)
        setStepperValue(stepper.integerValue)
    }
    
    func setStepperValue(_ integerValue: Int) {
        
        // sync stepper and text field values
        let previousValue = self.integerValue
        stepper.objectValue = integerValue
        objectValue = integerValue
        
        // return if the value didn't change
        if self.integerValue == previousValue { return }
        
        // update binding, if valid
        if let bindingInfo = infoForBinding(NSBindingName.value), let observedKeyPath = bindingInfo[NSBindingInfoKey.observedKeyPath] as? String {
            let observedObject = bindingInfo[NSBindingInfoKey.observedObject] as? NSObject
//            let observedAssembly = bindingInfo[NSObservedObjectKey] as? ROMWindowController
            observedObject?.setValue(objectValue, forKeyPath: observedKeyPath)
        }
    }
    
    override var acceptsFirstResponder: Bool { return true }
    
    override func resetCursorRects() {
        let cursorRect = NSMakeRect(0, 0, frame.width - stepperWidth, frame.height)
        addCursorRect(cursorRect, cursor: NSCursor.iBeam)
    }
    
    override func resizeSubviews(withOldSize oldSize: NSSize) {
        stepper.frame = NSMakeRect(frame.width - stepperWidth, 0, stepperWidth, frame.height)
    }
    
    // TODO: resize, menu
    
    func control(_ control: NSControl, textView: NSTextView, doCommandBy commandSelector: Selector) -> Bool {
        
        // TODO: get the stepper button to highlight
        
        switch commandSelector {
        case #selector(NSResponder.moveUp):
//            integerValue += smallIncrement
            setStepperValue(integerValue + smallIncrement)
//            stepper.intValue = intValue
            return true

        case #selector(NSResponder.moveDown):
//            integerValue -= smallIncrement
            setStepperValue(integerValue - smallIncrement)
//            stepper.intValue = intValue
            return true

        case #selector(NSResponder.moveToBeginningOfDocument(_:)):
//            integerValue += bigIncrement
            setStepperValue(integerValue + bigIncrement)
//            stepper.intValue = intValue
            return true

        case #selector(NSResponder.moveToEndOfDocument(_:)):
//            integerValue -= bigIncrement
            setStepperValue(integerValue - bigIncrement)
//            stepper.intValue = intValue
            return true

        default:
            return false
        }
    }
}
